# -*- coding: utf-8 -*-
"""geog527train.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1N4Hv5AG0Fo_3H9eFjMEvNC1uOqBWTbCb

# GEOG 527 Final Project
### Adam J. Stewart

## Setup

First, the `torchsat` repository is uploaded onto Google Drive. Then, Google Drive is mounted in Google Colab.
"""

from google.colab import drive
drive.mount('/content/drive', force_remount=True)

"""Some PyTorch tutorial code is added to the current directory. See https://pytorch.org/tutorials/intermediate/torchvision_tutorial.html for the original tutorial this is based off of."""

# Commented out IPython magic to ensure Python compatibility.
# %%shell
# 
# git clone https://github.com/pytorch/vision.git
# cd vision
# 
# cp references/detection/utils.py ../
# cp references/detection/transforms.py ../
# cp references/detection/coco_eval.py ../
# cp references/detection/engine.py ../
# cp references/detection/coco_utils.py ../

"""Next, the dependencies are installed."""

!pip install Pillow pycocotools rarfile

"""Finally, `torchsat` is added to the `sys.path` so it can be imported."""

import sys
sys.path.append('/content/drive/MyDrive/satellite')

"""## Dataset

First, we define a set of transforms to apply to the image and target.
"""

import transforms as T
import coco_utils as CU
import torchsat.transforms as TST

train_transforms = T.Compose([
    CU.ConvertCocoPolysToMask(),
    T.ToTensor(),
    TST.RandomHorizontalFlip(0.5),
    TST.RandomVerticalFlip(0.5),
])

test_transforms = T.Compose([
    CU.ConvertCocoPolysToMask(),
    T.ToTensor(),
])

"""The next step is to initialize the dataset. The dataset is automatically downloaded and checksummed by the `VHR10` dataset class. The dataset is then randomly split into training and testing sets."""

import torch
from torchsat.datasets import VHR10

dataset = VHR10(root='data', transforms=train_transforms, download=True)

train_dataset, test_dataset = torch.utils.data.random_split(dataset, [520, 130], generator=torch.Generator().manual_seed(1))

"""Finally, data loaders are created for each dataset."""

import utils

train_loader = torch.utils.data.DataLoader(
    train_dataset, batch_size=4, shuffle=True, num_workers=2,
    collate_fn=utils.collate_fn)

test_loader = torch.utils.data.DataLoader(
    test_dataset, batch_size=4, shuffle=False, num_workers=2,
    collate_fn=utils.collate_fn)

"""## Model

For this project, I'll be using [Mask R-CNN](https://arxiv.org/abs/1703.06870), which is based on top of [Faster R-CNN](https://arxiv.org/abs/1506.01497). Faster R-CNN is a model that predicts both bounding boxes and class scores for potential objects in the image.

![Faster R-CNN](https://raw.githubusercontent.com/pytorch/vision/temp-tutorial/tutorials/tv_image03.png)

Mask R-CNN adds an extra branch into Faster R-CNN, which also predicts segmentation masks for each instance.

![Mask R-CNN](https://raw.githubusercontent.com/pytorch/vision/temp-tutorial/tutorials/tv_image04.png)

We'll start with a Mask R-CNN model pre-trained on COCO.
"""

import torchvision
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor

# Dataset has 10 classes + background
num_classes = 11
      
# Load an instance segmentation model pre-trained on COCO
model = torchvision.models.detection.maskrcnn_resnet50_fpn(pretrained=True, rpn_nms_thresh=0.5, box_nms_thresh=0.3)

# Get the number of input features for the classifier
in_features = model.roi_heads.box_predictor.cls_score.in_features
# Replace the pre-trained head with a new one
model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

# Get the number of input features for the mask classifier
in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
hidden_layer = 256
# Replace the mask predictor with a new one
model.roi_heads.mask_predictor = MaskRCNNPredictor(
    in_features_mask, hidden_layer, num_classes)

"""We also want to move the model to the GPU for faster training and evaluation."""

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

model = model.to(device)

"""## Training

First, we define an optimizer and learning rate scheduler.
"""

params = [p for p in model.parameters() if p.requires_grad]
optimizer = torch.optim.SGD(
    params, lr=0.005, momentum=0.9, weight_decay=0.0005)

lr_scheduler = torch.optim.lr_scheduler.StepLR(
    optimizer, step_size=3, gamma=0.1)

"""Then we train the model for 10 epochs, evaluating at the end of every epoch."""

from engine import train_one_epoch, evaluate

num_epochs = 10

for epoch in range(num_epochs):
    # Train
    train_one_epoch(model, optimizer, train_loader, device, epoch, print_freq=10)
    lr_scheduler.step()

    # Need to disable random flips during evaluation
    # Work around bug in reference code: https://github.com/pytorch/vision/issues/3841
    train_dataset.dataset.transforms = test_transforms
    evaluate(model, train_loader, device=device)
    evaluate(model, test_loader, device=device)
    train_dataset.dataset.transforms = train_transforms

    # Save checkpoint
    torch.save(model.state_dict(), 'checkpoint.pth')
    !cp checkpoint.pth /content/drive/MyDrive/satellite/checkpoint.pth

"""## Evaluation

First, we'll load the model we checksummed earlier. This allows us to resume evaluation even if we lose our connection to Colab.
"""

model.load_state_dict(torch.load('/content/drive/MyDrive/satellite/checkpoint.pth'))
model = model.eval()

"""Next, we'll define a function that allows us to plot the predicted bounding boxes, masks, and confidence levels for each input image."""

from matplotlib import patches
import matplotlib.pyplot as plt
import numpy as np
from skimage.measure import find_contours
from PIL import Image


def plot(i: int) -> None:
    """Plot a particular image in the testing set with its predictions."""

    data, target = test_dataset[i]
    with torch.no_grad():
        prediction = model([data.to(device)])[0]

    image = data.mul(255).permute(1, 2, 0).byte().numpy()
    boxes = prediction['boxes'].cpu().numpy()
    labels = prediction['labels'].cpu().numpy()
    masks = [mask.squeeze().cpu().numpy() for mask in prediction['masks']]
    scores = prediction['scores'].cpu().numpy()

    N = len(scores)
    class_names = ['background', 'airplane', 'ship', 'storage tank', 'baseball diamond', 'tennis court', 'basketball court', 'ground track field', 'harbor', 'bridge', 'vehicle']

    # Display image
    fig, ax = plt.subplots(figsize=(24, 16))
    ax.imshow(image)

    cm = plt.get_cmap('gist_rainbow')
    for i in range(N):
        score = scores[i]
        if score < 0.5:
            continue

        class_num = labels[i]
        color = cm(class_num/num_classes)

        # Add bounding boxes
        x1, y1, x2, y2 = boxes[i]
        p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2, alpha=0.7, linestyle='dashed', edgecolor=color, facecolor='none')
        ax.add_patch(p)

        # Add labels
        label = class_names[class_num]
        caption = f'{label} {score:.3f}'
        ax.text(x1, y1 + 8, caption, color='white', size=11, backgroundcolor='none')

        # Add masks
        mask = masks[i]
        contours = find_contours(mask, 0.5)
        for verts in contours:
            verts = np.fliplr(verts)
            p = patches.Polygon(verts, facecolor=color, alpha=0.5, edgecolor='white')
            ax.add_patch(p)

    plt.axis('off')
    plt.show()

"""Finally, we'll display examples from each target class."""

plot(15)

plot(59)

plot(81)

plot(100)

plot(13)

plot(2)

plot(86)